#include "localsolver.h"
#include <fstream>
#include <iostream>
#include <json.hpp>
#include <map>
#include <sstream>
#include <vector>

using namespace localsolver;
using namespace std;
using json = nlohmann::json;

class Pooling {
public:
    // LocalSolver
    LocalSolver localsolver;

    // Instance data
    int nbComponents;
    int nbProducts;
    int nbPools;
    int nbAttributes;

    vector<double> componentPrices;
    vector<double> componentSupplies;
    vector<vector<double>> componentQualities;
    vector<string> componentNames;
    map<string, int> componentIdx;

    vector<double> productCapacities;
    vector<double> productPrices;
    vector<double> demand;
    vector<vector<double>> maxTolerance;
    vector<vector<double>> minTolerance;
    vector<string> productNames;
    map<string, int> productIdx;

    vector<double> poolCapacities;
    vector<string> poolNames;
    map<string, int> poolIdx;

    vector<vector<double>> upperBoundComponentToProduct;
    vector<vector<double>> costComponentToProduct;

    vector<vector<double>> upperBoundFractionComponentToPool;
    vector<vector<double>> costComponentToPool;

    vector<vector<double>> upperBoundPoolToProduct;
    vector<vector<double>> costPoolToProduct;

    vector<vector<vector<LSExpression>>> flowComponentToProductByPool;

    // Decision variables
    vector<vector<LSExpression>> flowComponentToProduct;
    vector<vector<LSExpression>> flowPoolToProduct;
    vector<vector<LSExpression>> fractionComponentToPool;

    // Objective value
    LSExpression profit;

    /* Read instance data */
    void readInstance(const string& fileName) {
        ifstream infile;
        infile.exceptions(ifstream::failbit | ifstream::badbit);
        infile.open(fileName.c_str());

        json problem;
        problem << infile;
        infile.close();

        nbComponents = problem["components"].size();
        nbProducts = problem["products"].size();
        nbPools = problem["pool_size"].size();
        nbAttributes = problem["components"][0]["quality"].size();

        // Components
        componentPrices.resize(nbComponents);
        componentSupplies.resize(nbComponents);
        componentQualities.resize(nbComponents);
        componentNames.resize(nbComponents);

        for (int c = 0; c < nbComponents; ++c) {
            componentPrices[c] = problem["components"][c]["price"];
            componentSupplies[c] = problem["components"][c]["upper"];
            componentNames[c] = problem["components"][c]["name"].get<string>();
            componentIdx[componentNames[c]] = c;
            int currentAttributeIdx = 0;
            componentQualities[c].resize(nbAttributes);
            json qualityComponent = problem["components"][c]["quality"];
            for (json::iterator it = qualityComponent.begin(); it != qualityComponent.end(); ++it) {
                componentQualities[c][currentAttributeIdx] = it.value();
                currentAttributeIdx++;
            }
        }

        // Final products (blendings)
        productCapacities.resize(nbProducts);
        productPrices.resize(nbProducts);
        demand.resize(nbProducts);
        maxTolerance.resize(nbProducts);
        minTolerance.resize(nbProducts);
        productNames.resize(nbProducts);

        for (int p = 0; p < nbProducts; ++p) {
            productCapacities[p] = problem["products"][p]["upper"];
            productPrices[p] = problem["products"][p]["price"];
            demand[p] = problem["products"][p]["lower"];
            productNames[p] = problem["products"][p]["name"].get<string>();
            productIdx[productNames[p]] = p;

            maxTolerance[p].resize(nbAttributes);
            int currentAttributeIdx = 0;
            json qualityUpperProduct = problem["products"][p]["quality_upper"];
            for (json::iterator it = qualityUpperProduct.begin(); it != qualityUpperProduct.end(); ++it) {
                maxTolerance[p][currentAttributeIdx] = it.value();
                currentAttributeIdx++;
            }

            minTolerance[p].resize(nbAttributes);
            currentAttributeIdx = 0;
            json qualityLowerProduct = problem["products"][p]["quality_lower"];
            bool hasLower = (qualityLowerProduct != NULL);
            for (json::iterator it = qualityLowerProduct.begin(); it != qualityLowerProduct.end(); ++it) {
                if (hasLower)
                    minTolerance[p][currentAttributeIdx] = it.value();
                else
                    minTolerance[p][currentAttributeIdx] = 0;
                currentAttributeIdx++;
            }
        }

        // Intermediate pools
        poolNames.resize(nbPools);
        poolCapacities.resize(nbPools);
        int currentPoolIdx = 0;
        for (json::iterator it = problem["pool_size"].begin(); it != problem["pool_size"].end(); ++it) {
            poolCapacities[currentPoolIdx] = it.value();
            poolNames[currentPoolIdx] = it.key();
            poolIdx[poolNames[currentPoolIdx]] = currentPoolIdx;
            currentPoolIdx++;
        }

        // Flow graph

        // Edges from the components to the products
        upperBoundComponentToProduct.resize(nbComponents);
        costComponentToProduct.resize(nbComponents);
        flowComponentToProduct.resize(nbComponents);

        // Edges from the components to the pools
        upperBoundFractionComponentToPool.resize(nbComponents);
        costComponentToPool.resize(nbComponents);
        fractionComponentToPool.resize(nbComponents);

        // Edges from the pools to the products
        upperBoundPoolToProduct.resize(nbPools);
        costPoolToProduct.resize(nbPools);
        flowPoolToProduct.resize(nbPools);

        // Flow from the components to the products and passing by the pools
        flowComponentToProductByPool.resize(nbComponents);

        for (int c = 0; c < nbComponents; ++c) {
            upperBoundComponentToProduct[c].resize(nbProducts);
            costComponentToProduct[c].resize(nbProducts);
            flowComponentToProduct[c].resize(nbProducts);

            upperBoundFractionComponentToPool[c].resize(nbPools);
            costComponentToPool[c].resize(nbPools);
            fractionComponentToPool[c].resize(nbPools);

            flowComponentToProductByPool[c].resize(nbPools);
            for (int o = 0; o < nbPools; ++o)
                flowComponentToProductByPool[c][o].resize(nbProducts);
        }

        for (int o = 0; o < nbPools; ++o) {
            upperBoundPoolToProduct[o].resize(nbProducts);
            costPoolToProduct[o].resize(nbProducts);
            flowPoolToProduct[o].resize(nbProducts);
        }

        // Bound and cost on the edges
        for (json& edge : problem["component_to_product_bound"]) {
            upperBoundComponentToProduct[componentIdx[edge["component"]]][productIdx[edge["product"]]] = edge["bound"];
            if (edge.size() > 3)
                costComponentToProduct[componentIdx[edge["component"]]][productIdx[edge["product"]]] = edge["cost"];
        }

        for (json& edge : problem["component_to_pool_fraction"]) {
            upperBoundFractionComponentToPool[componentIdx[edge["component"]]][poolIdx[edge["pool"]]] =
                edge["fraction"];
            if (edge.size() > 3)
                costComponentToPool[componentIdx[edge["component"]]][poolIdx[edge["pool"]]] = edge["cost"];
        }

        for (json& edge : problem["pool_to_product_bound"]) {
            upperBoundPoolToProduct[poolIdx[edge["pool"]]][productIdx[edge["product"]]] = edge["bound"];
            if (edge.size() > 3)
                costPoolToProduct[poolIdx[edge["pool"]]][productIdx[edge["product"]]] = edge["cost"];
        }
    }

    // Declare the optimization model
    void solve(int limit) {
        LSModel model = localsolver.getModel();

        // Decision variables

        // Flow from the components to the products
        for (int c = 0; c < nbComponents; ++c) {
            for (int p = 0; p < nbProducts; ++p)
                flowComponentToProduct[c][p] = model.floatVar(0, upperBoundComponentToProduct[c][p]);
        }
        // Fraction of the total flow in pool o coming from the component c
        for (int c = 0; c < nbComponents; ++c) {
            for (int o = 0; o < nbPools; ++o)
                fractionComponentToPool[c][o] = model.floatVar(0, upperBoundFractionComponentToPool[c][o]);
        }
        // Flow from the pools to the products
        for (int o = 0; o < nbPools; ++o) {
            for (int p = 0; p < nbProducts; ++p)
                flowPoolToProduct[o][p] = model.floatVar(0, upperBoundPoolToProduct[o][p]);
        }

        // Flow from the components to the products and passing by the pools
        for (int c = 0; c < nbComponents; ++c) {
            for (int o = 0; o < nbPools; ++o) {
                for (int p = 0; p < nbProducts; ++p)
                    flowComponentToProductByPool[c][o][p] = fractionComponentToPool[c][o] * flowPoolToProduct[o][p];
            }
        }

        // Constraints

        // Proportion
        for (int o = 0; o < nbPools; ++o) {
            LSExpression proportion = model.sum();
            for (int c = 0; c < nbComponents; ++c)
                proportion.addOperand(fractionComponentToPool[c][o]);
            model.constraint(proportion == 1);
        }

        // Component supply
        for (int c = 0; c < nbComponents; ++c) {
            LSExpression flowToProducts = model.sum(flowComponentToProduct[c].begin(), flowComponentToProduct[c].end());
            LSExpression flowToPools = model.sum();
            for (int o = 0; o < nbPools; ++o) {
                flowToPools.addOperand(
                    model.sum(flowComponentToProductByPool[c][o].begin(), flowComponentToProductByPool[c][o].end()));
            }
            LSExpression totalOutFlow = model.sum(flowToPools, flowToProducts);
            model.constraint(totalOutFlow <= componentSupplies[c]);
        }

        // Pool capacity (bounds on edges)
        for (int c = 0; c < nbComponents; ++c) {
            for (int o = 0; o < nbPools; ++o) {
                LSExpression flowComponentToPool =
                    model.sum(flowComponentToProductByPool[c][o].begin(), flowComponentToProductByPool[c][o].end());
                LSExpression edgeCapacity = model.prod(poolCapacities[o], fractionComponentToPool[c][o]);
                model.constraint(flowComponentToPool <= edgeCapacity);
            }
        }

        // Product capacity
        for (int p = 0; p < nbProducts; ++p) {
            LSExpression flowFromPools = model.sum();
            for (int o = 0; o < nbPools; ++o)
                flowFromPools.addOperand(flowPoolToProduct[o][p]);
            LSExpression flowFromComponents = model.sum();
            for (int c = 0; c < nbComponents; ++c)
                flowFromComponents.addOperand(flowComponentToProduct[c][p]);
            LSExpression totalInFlow = model.sum(flowFromComponents, flowFromPools);
            model.constraint(totalInFlow <= productCapacities[p]);
            model.constraint(totalInFlow >= demand[p]);
        }

        // Product tolerance
        for (int p = 0; p < nbProducts; ++p) {
            for (int k = 0; k < nbAttributes; ++k) {

                // Attribute from the components
                LSExpression attributeFromComponents = model.sum();
                for (int c = 0; c < nbComponents; ++c)
                    attributeFromComponents.addOperand(componentQualities[c][k] * flowComponentToProduct[c][p]);

                // Attribute from the pools
                LSExpression attributeFromPools = model.sum();
                for (int c = 0; c < nbComponents; ++c) {
                    for (int o = 0; o < nbPools; ++o)
                        attributeFromPools.addOperand(componentQualities[c][k] * flowComponentToProductByPool[c][o][p]);
                }

                // Total flow in the blending
                LSExpression totalFlowIn = model.sum();
                for (int c = 0; c < nbComponents; ++c)
                    totalFlowIn.addOperand(flowComponentToProduct[c][p]);
                for (int o = 0; o < nbPools; ++o)
                    totalFlowIn.addOperand(flowPoolToProduct[o][p]);

                LSExpression totalAttributeIn = model.sum(attributeFromComponents, attributeFromPools);
                model.constraint(totalAttributeIn >= minTolerance[p][k] * totalFlowIn);
                model.constraint(totalAttributeIn <= maxTolerance[p][k] * totalFlowIn);
            }
        }

        // Objective function

        // Cost of the flows from the components directly to the products
        LSExpression directFlowCost = model.sum();

        // Cost of the flows from the components to the products passing by the pools
        LSExpression undirectFlowCost = model.sum();
        for (int c = 0; c < nbComponents; ++c) {
            for (int p = 0; p < nbProducts; ++p) {
                directFlowCost.addOperand(costComponentToProduct[c][p] * flowComponentToProduct[c][p]);
                for (int o = 0; o < nbPools; ++o) {
                    undirectFlowCost.addOperand((costComponentToPool[c][o] + costPoolToProduct[o][p]) *
                                                flowComponentToProductByPool[c][o][p]);
                }
            }
        }

        // Gain of selling the final products
        LSExpression productsGain = model.sum();
        for (int p = 0; p < nbProducts; ++p) {
            LSExpression totalFlowInProduct = model.sum();
            for (int c = 0; c < nbComponents; ++c)
                totalFlowInProduct.addOperand(flowComponentToProduct[c][p]);
            for (int o = 0; o < nbPools; ++o)
                totalFlowInProduct.addOperand(flowPoolToProduct[o][p]);
            productsGain.addOperand(productPrices[p] * totalFlowInProduct);
        }

        // Cost of buying the components
        LSExpression componentsCost = model.sum();
        for (int c = 0; c < nbComponents; ++c) {
            LSExpression totalFlowInComponent = model.sum();
            for (int p = 0; p < nbProducts; ++p) {
                totalFlowInComponent.addOperand(flowComponentToProduct[c][p]);
                for (int o = 0; o < nbPools; ++o)
                    totalFlowInComponent.addOperand(fractionComponentToPool[c][o] * flowPoolToProduct[o][p]);
            }
            componentsCost.addOperand(componentPrices[c] * totalFlowInComponent);
        }

        LSExpression totalCost = model.sum(componentsCost, directFlowCost, undirectFlowCost);
        profit = model.sub(productsGain, totalCost);

        // Maximize the total profit
        model.maximize(profit);
        model.close();

        // Parameterize the solver
        localsolver.getParam().setTimeLimit(limit);

        localsolver.solve();
    }

    /* Write the solution */
    void writeSolution(const string& fileName) {
        ofstream outfile;
        outfile.exceptions(ofstream::failbit | ofstream::badbit);
        outfile.open(fileName.c_str());

        vector<json> component_to_product;
        vector<json> component_to_pool_fraction;
        vector<json> pool_to_product;
        component_to_product.resize(nbComponents * nbProducts);
        component_to_pool_fraction.resize(nbComponents * nbPools);
        pool_to_product.resize(nbPools * nbProducts);

        // Solution flows from the components to the products
        for (int c = 0; c < nbComponents; ++c) {
            for (int p = 0; p < nbProducts; ++p) {
                component_to_product[c * nbProducts + p] = {{"component", componentNames[c]},
                                                            {"product", productNames[p]},
                                                            {"flow", flowComponentToProduct[c][p].getDoubleValue()}};
            }
        }
        // Solution fraction of the inflow at pool o coming from the component c
        for (int c = 0; c < nbComponents; ++c) {
            for (int o = 0; o < nbPools; ++o) {
                component_to_pool_fraction[c * nbPools + o] = {
                    {"component", componentNames[c]},
                    {"pool", poolNames[o]},
                    {"flow", fractionComponentToPool[c][o].getDoubleValue()}};
            }
        }
        // Solution flows from the pools to the products
        for (int o = 0; o < nbPools; ++o) {
            for (int p = 0; p < nbProducts; ++p) {
                pool_to_product[o * nbProducts + p] = {{"pool", poolNames[o]},
                                                       {"product", productNames[p]},
                                                       {"flow", flowPoolToProduct[o][p].getDoubleValue()}};
            }
        }
        json solution = {{"objective", profit.getDoubleValue()},
                         {"solution",
                          {{"component_to_pool_fraction", component_to_pool_fraction},
                           {"component_to_product", component_to_product},
                           {"pool_to_product", pool_to_product}}}};

        outfile << solution;
    }
};

int main(int argc, char** argv) {
    if (argc < 2) {
        cerr << "Usage: pooling inputFile [outputFile] [timeLimit] " << endl;
        return 1;
    }

    const char* instanceFile = argv[1];
    const char* solFile = argc > 2 ? argv[2] : NULL;
    const char* strTimeLimit = argc > 3 ? argv[3] : "20";

    try {
        Pooling model;
        model.readInstance(instanceFile);
        model.solve(atoi(strTimeLimit));
        if (solFile != NULL)
            model.writeSolution(solFile);
        return 0;
    } catch (const exception& e) {
        cerr << "An error occurred: " << e.what() << endl;
        return 1;
    }
}
